// This file will be copied to https://www.open3d.org/docs/versions.js
// by the documentation server.
document.write('\
    <!-- ac-hoc table style, put in in proper css file later -->\
    <style>\
    .version-table {\
        width: 100%;\
        table-layout: fixed;\
    }\
    table.version-table td {\
        border: 1px solid #555555;\
        text-align: left;\
    }\
    </style>\
    <table class="version-table">\
        <tr>\
            <td><a href="https://www.open3d.org/docs/latest">main</a></td>\
            <td><a href="https://www.open3d.org/docs/latest/cpp_api">main C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/release">0.19.0 (release)</a></td>\
            <td><a href="https://www.open3d.org/docs/release/cpp_api">0.19.0 C++ (release)</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/release">0.18.0</a></td>\
            <td><a href="https://www.open3d.org/docs/release/cpp_api">0.18.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.17.0">0.17.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.17.0/cpp_api">0.17.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.16.0">0.16.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.16.0/cpp_api">0.16.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.15.1">0.15.1</a></td>\
            <td><a href="https://www.open3d.org/docs/0.15.1/cpp_api">0.15.1 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.14.1">0.14.1</a></td>\
            <td><a href="https://www.open3d.org/docs/0.14.1/cpp_api">0.14.1 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.13.0">0.13.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.13.0/cpp_api">0.13.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.12.0">0.12.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.12.0/cpp_api">0.12.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.11.1">0.11.1</a></td>\
            <td><a href="https://www.open3d.org/docs/0.11.1/cpp_api">0.11.1 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.11.0">0.11.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.11.0/cpp_api">0.11.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.10.0">0.10.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.10.0/cpp_api">0.10.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.9.0">0.9.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.9.0/cpp_api">0.9.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.8.0">0.8.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.8.0/cpp_api">0.8.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.7.0">0.7.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.7.0/cpp_api">0.7.0 C++</a></td>\
        </tr>\
        <tr>\
            <td><a href="https://www.open3d.org/docs/0.6.0">0.6.0</a></td>\
            <td><a href="https://www.open3d.org/docs/0.6.0/cpp_api">0.6.0 C++</a></td>\
        </tr>\
    </table>\
');
