// ----------------------------------------------------------------------------
// -                        Open3D: www.open3d.org                            -
// ----------------------------------------------------------------------------
// Copyright (c) 2018-2024 www.open3d.org
// SPDX-License-Identifier: MIT
// ----------------------------------------------------------------------------

#include "tests/Tests.h"

namespace open3d {
namespace tests {

TEST(FilePLY, DISABLED_ReadVertexCallback) { NotImplemented(); }

TEST(FilePLY, DISABLED_AdvanceConsoleProgress) { NotImplemented(); }

TEST(FilePLY, DISABLED_ReadNormalCallback) { NotImplemented(); }

TEST(FilePLY, DISABLED_ReadColorCallback) { NotImplemented(); }

TEST(FilePLY, DISABLED_ReadFaceCallBack) { NotImplemented(); }

TEST(FilePLY, DISABLED_ReadPointCloudFromPLY) { NotImplemented(); }

TEST(FilePLY, DISABLED_WritePointCloudToPLY) { NotImplemented(); }

TEST(FilePLY, DISABLED_ReadTriangleMeshFromPLY) { NotImplemented(); }

TEST(FilePLY, DISABLED_WriteTriangleMeshToPLY) { NotImplemented(); }

TEST(FilePLY, DISABLED_ResetConsoleProgress) { NotImplemented(); }

}  // namespace tests
}  // namespace open3d
