// ----------------------------------------------------------------------------
// -                        Open3D: www.open3d.org                            -
// ----------------------------------------------------------------------------
// Copyright (c) 2018-2024 www.open3d.org
// SPDX-License-Identifier: MIT
// ----------------------------------------------------------------------------

#include "tests/Tests.h"

namespace open3d {
namespace tests {

TEST(GlobalOptimization, DISABLED_Constructor) { NotImplemented(); }

TEST(GlobalOptimization, DISABLED_MemberData) { NotImplemented(); }

TEST(GlobalOptimization, DISABLED_GlobalOptimizationLevenbergMarquardt) {
    NotImplemented();
}

TEST(GlobalOptimization, DISABLED_GlobalOptimizationConvergenceCriteria) {
    NotImplemented();
}

TEST(GlobalOptimization, DISABLED_GlobalOptimizationOption) {
    NotImplemented();
}

TEST(GlobalOptimization, DISABLED_CreatePoseGraphWithoutInvalidEdges) {
    NotImplemented();
}

}  // namespace tests
}  // namespace open3d
