nav {
  background-color: #333;
  overflow: hidden;
}
nav a {
  float: left;
  display: block;
  color: #f2f2f2;
  text-align: center;
  padding: 10px 10px;
}
nav a:hover {
  background-color: #ddd;
  color: black;
}
nav a.active {
  background-color: #4caf50;
  color: white;
}
nav a.active:hover {
  background-color: #4caf50;
  color: black;
}
