// ----------------------------------------------------------------------------
// -                        Open3D: www.open3d.org                            -
// ----------------------------------------------------------------------------
// Copyright (c) 2018-2024 www.open3d.org
// SPDX-License-Identifier: MIT
// ----------------------------------------------------------------------------

#include "tests/Tests.h"

namespace open3d {
namespace tests {

TEST(FilePCD, DISABLED_CheckHeader) { NotImplemented(); }

TEST(FilePCD, DISABLED_ReadPCDHeader) { NotImplemented(); }

TEST(FilePCD, DISABLED_SplitString) { NotImplemented(); }

TEST(FilePCD, DISABLED_UnpackBinaryPCDElement) { NotImplemented(); }

TEST(FilePCD, DISABLED_UnpackBinaryPCDColor) { NotImplemented(); }

TEST(FilePCD, DISABLED_UnpackASCIIPCDElement) { NotImplemented(); }

TEST(FilePCD, DISABLED_UnpackASCIIPCDColor) { NotImplemented(); }

TEST(FilePCD, DISABLED_ReadPCDData) { NotImplemented(); }

TEST(FilePCD, DISABLED_RemoveNanData) { NotImplemented(); }

TEST(FilePCD, DISABLED_GenerateHeader) { NotImplemented(); }

TEST(FilePCD, DISABLED_WritePCDHeader) { NotImplemented(); }

TEST(FilePCD, DISABLED_ConvertRGBToFloat) { NotImplemented(); }

TEST(FilePCD, DISABLED_WritePCDData) { NotImplemented(); }

TEST(FilePCD, DISABLED_ReadPointCloudFromPCD) { NotImplemented(); }

TEST(FilePCD, DISABLED_WritePointCloudToPCD) { NotImplemented(); }

}  // namespace tests
}  // namespace open3d
