// ----------------------------------------------------------------------------
// -                        Open3D: www.open3d.org                            -
// ----------------------------------------------------------------------------
// Copyright (c) 2018-2024 www.open3d.org
// SPDX-License-Identifier: MIT
// ----------------------------------------------------------------------------

#include "tests/Tests.h"

namespace open3d {
namespace tests {

TEST(Timer, DISABLED_Constructor) { NotImplemented(); }

TEST(Timer, DISABLED_Destructor) { NotImplemented(); }

TEST(Timer, DISABLED_MemberData) { NotImplemented(); }

TEST(Timer, DISABLED_GetSystemTimeInMilliseconds) { NotImplemented(); }

TEST(Timer, DISABLED_Start) { NotImplemented(); }

TEST(Timer, DISABLED_Stop) { NotImplemented(); }

TEST(Timer, DISABLED_Print) { NotImplemented(); }

TEST(Timer, DISABLED_GetDuration) { NotImplemented(); }

TEST(Timer, DISABLED_ScopeTimer) { NotImplemented(); }

TEST(Timer, DISABLED_FPSTimer) { NotImplemented(); }

TEST(Timer, DISABLED_Signal) { NotImplemented(); }

}  // namespace tests
}  // namespace open3d
