@import url("/nav.css");
table {
  margin: 0 auto;
}
h2 {
  margin: 0 auto;
  text-align: center;
}
footer {
  text-align: center;
}
div {
  margin: 0 auto;
  text-align: center;
}
button {
  margin: 0 auto;
  display: block;
}
video {
  margin: auto;
  left: 0;
  right: 0;
  position: relative;
  background: grey;
}
canvas {
  position: absolute;
  margin: auto;
  left: 0;
  right: 0;
  z-index: 999;
}
