// ----------------------------------------------------------------------------
// -                        Open3D: www.open3d.org                            -
// ----------------------------------------------------------------------------
// Copyright (c) 2018-2024 www.open3d.org
// SPDX-License-Identifier: MIT
// ----------------------------------------------------------------------------

#include "tests/Tests.h"

namespace open3d {
namespace tests {

TEST(ScalableTSDFVolume, DISABLED_VolumeUnit) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_Constructor) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_Destructor) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_MemberData) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_Reset) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_Integrate) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_ExtractPointCloud) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_ExtractTriangleMesh) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_ExtractVoxelPointCloud) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_LocateVolumeUnit) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_OpenVolumeUnit) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_GetNormalAt) { NotImplemented(); }

TEST(ScalableTSDFVolume, DISABLED_GetTSDFAt) { NotImplemented(); }

}  // namespace tests
}  // namespace open3d
